# Copyright 2014, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from a10_neutron_lbaas.plumbing.wrappers import AcosWrapper
from a10_neutron_lbaas.plumbing.wrappers import NeutronDbWrapper

from a10_neutron_lbaas.plumbing import simple


LOG = log.getLogger(__name__)


VLAN_LOG_FMT = "VLAN:{0} Project:{1} Network:{2}"


class VlanPortBindingPlumbingHooks(simple.PlumbingHooks):
    def after_member_create(self, a10_context, os_context, member):
        pass

    def after_member_update(self, a10_context, os_context, member):
        pass

    def pre_vip_create_v1(self, a10_context, os_context, vip):
        # msg = "PRE_CREATE_V1 hook firing", vip, dir(vip)
        # LOG.debug(msg)
        # cidr = self._get_vip_cidr(a10_context, os_context, vip, v1=True)
        # self._create_nat_pool(a10_context, os_context, vip, v1=True)
        # return cidr
        pass

    def pre_vip_create_v2(self, a10_context, os_context, vip):
        # db = NeutronDbWrapper(os_context.session)
        # acos = AcosWrapper(a10_context.client)
        # config = a10_context.a10_driver.config
        # self._create_nat_pool(a10_context, os_context, vip, v1=False)
        # return self._get_vip_cidr(a10_context, os_context, vip)
        pass

    def after_vip_create(self, a10_context, os_context, vip):
        # Get the IDs of all the things we need.
        db = NeutronDbWrapper(os_context.session)
        acos = AcosWrapper(a10_context.client)
        config = a10_context.a10_driver.config
        # v2 obj model
        if hasattr(vip, "vip_port"):
            vip_port = vip.vip_port
            subnet_id = vip.vip_subnet_id
            subnet = db.get_subnet(subnet_id)
            port_id = vip_port.id
            network_id = vip_port.network_id
            tenant_id = vip.tenant_id
            vip_id = vip.id

        # v1 obj model
        else:
            subnet_id = vip["subnet_id"]
            port_id = vip["port_id"]
            subnet = db.get_subnet(subnet_id)
            network_id = subnet["network_id"]
            tenant_id = vip["tenant_id"]
            vip_id = vip["id"]

        try:
            binding_level = config.get("vlan_binding_level")
        except Exception as ex:
            # If binding_level isn't configured, there is a lot that can go wrong.
            # Log an error and bail.
            # TODO(mdurrant) Narrow exception handling retrieving binding level
            raise ex

        # Get {network_id} attached to the port.network_id
        # Get the segment associated to {network_id} with the network matching our criteria
        segment = db.get_segment(port_id, binding_level)
        # Get {vlan_id} from the segment
        vlan_id = segment.segmentation_id

        ve = None
        # Is there one assigned to the segment?
        if not vlan_id:
            LOG.info(
                "No VLAN ID for port {0} on segment {1}. Exiting hook.".format(port_id, segment.id))
            return
        # Format the string we're going to use a gazillion times for logging.

        # Get the associated VE
        ve = acos.get_ve(vlan_id)
        # If the VE already exists, the VLAN already exists and in most cases we're done.
        # Log that this happened.
        # TODO(mdurrant) - Find a way to check if it's in another partition and gripe loudly.
        if ve:
            LOG.info("Configuration for {0} previously created".format(vlan_id))
            return

        # If DHCP, we can configure the interface with DHCP
        use_dhcp = config.get("plumb_vlan_dhcp")

        # Create the VLAN with configured interfaces
        interfaces = config.get("vlan_interfaces")
        acos.create_vlan(vlan_id, interfaces)

        # Get the MAC of the VE so we can create the port.
        ve_pre = acos.get_ve(vlan_id)
        ve_mac = ve_pre.get("ve", {}).get("oper", {}).get("mac")
        if not ve_mac:
            LOG.error("Could not retrieve VE MAC, port binding operation failed.")
            return
        ve_mac = self._format_mac(ve_mac)

        # Create a neutron port for VE interface using returned VE MAC
        ve_nport = db.create_port(network_id, tenant_id, ve_mac, network_id)
        # Get the vips port from neutron
        vip_port = db.get_port(port_id)
        vip_mac = vip_port["mac_address"]
        acos.update_vip(vip_id, vip_mac, vlan_id)
        LOG.info("Updated VIP {0} with mac {1} vlan {2}".format(vip_id, vip_mac, vlan_id))
        # TODO(mdurrant) - Should host be blank
        ve_portbinding = db.create_port_binding(ve_nport.id, "")
        assert(ve_portbinding is not None)

        # Set defaults
        ve_ip, ve_mask, ve_port = None, None, ve_nport.id

        if not use_dhcp:
            # Allocate IP from the subnet.
            ve_ip, ve_mask, ve_port = db.allocate_ip_for_subnet(subnet_id, ve_mac, ve_nport.id)

        LOG.info("Created VLAN {0} with interfaces {1}".format(str(vlan_id), str(interfaces)))
        ve_dict = self._build_ve_dict(vlan_id, ve_ip, ve_mask, use_dhcp)

        # Try to create it. If an exception is raised, it's logged and we stop here.
        ve_created = acos.create_ve(ve_dict)
        if not ve_created:
            LOG.error("Exception creating VE interface for VLAN:{0}".format(vlan_id))
            return

        # Log the created VE for troubleshooting purposes.
        LOG.info("Created VE {0}", str(vlan_id))

        # Get the IP address of the port
        # if use_dhcp:
        #    LOG.info("VE {0}", str(ve_created))

        # Else, it already exists and we're good
        LOG.info("Configured {0} with interface IP: {1}".format(
            vlan_id, ve_ip))
        # fin.

    def after_vip_update(self, a10_context, os_context, vip):
        pass

    def partition_delete_last(self, client, openstack_context, name, lbaas_obj):
        # After partition delete, remove any neutron ports created by this hook owned by the tenant
        # Make sure the last object is an LB. The data model ensures this but ... just in case.
        # or type(lbaas_obj).__name__ != "LoadBalancer":
        if not lbaas_obj:
            LOG.info("No lbaas obj was set for cleanup, exiting cleanup")
            return

        db = NeutronDbWrapper(openstack_context.session)
        LOG.info("Cleanup ports for {0}".format(name))
        db.cleanup_vlan_ports(lbaas_obj)

    def _build_ve_dict(self, vlan_id, ip=None, mask=None, use_dhcp=None):
        rval = {"ifnum": vlan_id}
        if use_dhcp:
            rval["dhcp"] = use_dhcp
        if ip and mask:
            rval["ip_address"] = ip
            rval["ip_netmask"] = self._format_cidr(mask)

        return rval

    def _format_mac(self, raw_mac):
        raw_mac = raw_mac.replace(".", "")
        mac = ':'.join(a + b for a, b in zip(raw_mac[::2], raw_mac[1::2]))
        return mac

    def _format_cidr(self, cidr):
        # Takes "A.B.C.D/n" and returns the "/n" portion
        marker = "/"
        if cidr and cidr[0] != marker:
            beg = cidr.find(marker)
            cidr = cidr[beg:]
        return cidr

    def _get_vip_cidr(self, a10_context, os_context, vip, v1=False):
        db = NeutronDbWrapper(os_context.session)
        if v1 is True:
            sid = vip["subnet_id"]
        else:
            sid = vip.vip_subnet_id
        vip_subnet_info = db.get_subnet(sid)
        return self._format_cidr(vip_subnet_info.cidr)

    def _create_nat_pool(self, a10_context, os_context, vip, v1=False):
        msg = "CREATE_NAT", (v1 is True), v1, type(v1)
        LOG.debug(msg)
        acos = AcosWrapper(a10_context.client)
        vip_cidr = self._get_vip_cidr(a10_context, os_context, vip, v1=v1)
        if v1 is True:
            try:
                acos.create_nat_pool(vip['id'], vip['address'], vip['address'], vip_cidr)
            except Exception as ex:
                raise ex
        else:
            try:
                acos.create_nat_pool(vip.id, vip.vip_address, vip.vip_address, vip_cidr)
            except Exception as ex:
                raise ex



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from nose.plugins.attrib import attr

from a10_neutron_lbaas.tests.unit import test_base

from a10_neutron_lbaas.plumbing import portbinding_vlan

from neutron.db.models.segment import NetworkSegment
from neutron.db.models_v2 import Port
from neutron.plugins.ml2.models import PortBindingLevel

_SUBNET_ID = "mysubnet"
_PORT_ID = "portid"
_SEGMENT_ID = "mysegment"
_VLAN_ID = 2000
_NETWORK_ID = "mynetwork"
_LEVEL = 1


def raise_(exc):
    raise exc


class FakeModel(object):
    def __init__(self, **kwargs):
        self._dict = kwargs

    def __getattr__(self, key):
        return self._dict.get(key) or raise_(KeyError(key))


class FakeSession(object):
    _model_map = {
        PortBindingLevel: FakeModel(segment_id=_SEGMENT_ID, level=_LEVEL),
        NetworkSegment: FakeModel(id=_SEGMENT_ID, segmentation_id=_VLAN_ID),
        Port: FakeModel(id=_PORT_ID, subnet_id=_SUBNET_ID, network_id=_NETWORK_ID),
    }

    def __init__(self, *args, **kwargs):
        pass

    def query(self, model):
        self._rval = self._model_map[model]
        return self

    def filter_by(self, *args, **kwargs):
        return self

    def first(self):
        return self._rval


class FakeConfig(object):
    def __init__(self, *args, **kwargs):
        self._dict = kwargs

    def get(self, key):
        return self._dict.get(key)


@attr(db=True)
class TestVlanPortBindingPlumbing(test_base.UnitTestBase):
    version = "v2"

    def setUp(self, *args, **kwargs):
        super(TestVlanPortBindingPlumbing, self).setUp(*args, **kwargs)
        self._subnet_id = "mysubnet"
        self._port_id = "portid"
        self._segment_id = "mysegment"
        self._vlan_id = 2000
        self._network_id = "mynetwork"
        self._level = 1

        self._binding_level = mock.Mock(level=1, segment_id=self._segment_id)
        self._segment = mock.Mock(id=self._segment_id, segmentation_id=self._vlan_id)
        self._build_mocks()
        self.target = portbinding_vlan.VlanPortBindingPlumbingHooks(self._driver, self._devices)

    def test_select_device(self):
        a = self.target.select_device("first-token")
        self.target.select_device("second-token")
        self.assertEqual(a, self.target.select_device("first-token"))

    def test_after_vip_create(self):
        self.target.after_vip_create(self.a10_context, self.os_context, self._vip)
        self._client.vlan.create.assert_called_once_with(_VLAN_ID, mock.ANY, mock.ANY)

    def test_after_vip_create_ve_exists(self):
        self._client.interface.ve.get.return_value = {"ve": 2}
        self.target.after_vip_create(self.a10_context, self.os_context, self._vip)
        self._client.vlan.create.assert_not_called()

    def _build_mocks(self):
        # a10_context dependencies
        self._vip = FakeModel(vip_subnet_id="mysubnet",
                              vip_port=FakeModel(id=self._port_id, network_id=self._network_id))
        self._devices = {"a": {"host": "1.2.3.4", "api_version": "3.0"}}
        self._driver = mock.Mock()
        self._client = self._build_client()
        self._config = FakeConfig(
            use_database=False,
            vlan_interfaces={
                "tagged_trunks": [1, 2],
            },
            use_dhcp=False,
            vlan_binding_level=self._level
        )
        self._a10_driver = mock.Mock(config=self._config)
        self.a10_context = mock.Mock(a10_driver=self._a10_driver, client=self._client)
        self._session = self._build_session()
        self.os_context = mock.Mock(session=self._session, tenant_id="tenant")

    def _build_session(self, **kwargs):
        return FakeSession()

    def _build_client(self):
        rval = mock.Mock()
        rval.interface.ve.get_oper.return_value = None
        rval.interface.ve.get = lambda: {}

        return rval



# Copyright 2019, Omkar Telee (omkartelee01), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from a10_neutron_lbaas.tests.unit.v2 import fake_objs
from a10_neutron_lbaas.tests.unit.v2 import test_base


class TestL7Rule(test_base.HandlerTestBase):
    def test_create_l7Rule(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        policy = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                        "http//:google.com", 23)
        rule = fake_objs.FakeL7Rule(policy.id, 'FILE_TYPE',
                                    'STARTS_WITH', 'testkey', 'testvalue')
        policy.rules = []
        policy.rules.append(rule)
        rule.policy = policy
        self.a.l7rule.create(None, rule)
        sm = str(self.a.last_client.mock_calls)
        self.assertTrue('fake-l7policy-id-001' in sm)
        self.assertTrue('slb.aflex_policy.create' in sm)

    def test_update_l7Rule(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        policy = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                        "http//:google.com", 23)
        rule = fake_objs.FakeL7Rule(policy.id, 'FILE_TYPE',
                                    'STARTS_WITH', 'testkey', 'oldvalue')
        policy.rules = []
        policy.rules.append(rule)
        old_rule = rule
        old_rule.policy = policy
        rule = fake_objs.FakeL7Rule(policy.id, 'FILE_TYPE',
                                    'STARTS_WITH', 'testkey',
                                    'updatedtestvalue')
        policy.rules = []
        policy.rules.append(rule)
        rule.policy = policy
        self.a.l7rule.update(None, old_rule, rule)
        sm = str(self.a.last_client.mock_calls)

        self.assertTrue('fake-l7policy-id-001' in sm)
        self.assertTrue('slb.aflex_policy.create' in sm)
        self.assertTrue('updatedtestvalue' in sm)
        self.assertFalse('oldvalue' in sm)

    def test_delete_l7Rule(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        policy = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                        "http//:google.com", 23)
        rule = fake_objs.FakeL7Rule(policy.id, 'FILE_TYPE',
                                    'STARTS_WITH', 'testkey', 'testvalue')
        policy.rules = []
        policy.rules.append(rule)
        rule.policy = policy
        self.a.l7rule.delete(None, rule)
        sm = str(self.a.last_client.mock_calls)
        self.assertTrue('fake-l7policy-id-001' in sm)
        self.assertTrue('slb.aflex_policy.create' in sm)
        self.assertTrue('true' in sm)
        self.assertFalse('testvalue' in sm)

    def test_delete_l7Rule_from_multiple(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        policy = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                        "http//:google.com", 23)
        rule1 = fake_objs.FakeL7Rule(policy.id, 'FILE_TYPE',
                                     'STARTS_WITH', 'testkey', 'oldvalue')
        rule1.id = 'testrule1'
        policy.rules = []
        policy.rules.append(rule1)
        rule1.policy = policy
        self.a.l7rule.create(None, rule1)

        rule2 = fake_objs.FakeL7Rule(policy.id, 'FILE_TYPE',
                                     'STARTS_WITH', 'testkey',
                                     'updatedtestvalue')
        rule2.id = 'testrule2'
        policy.rules.append(rule2)
        rule2.policy = policy
        self.a.l7rule.create(None, rule2)
        self.a.l7rule.delete(None, rule2)
        sm = str(self.a.last_client.mock_calls)
        self.assertTrue('fake-l7policy-id-001' in sm)
        self.assertTrue('slb.aflex_policy.create' in sm)
        self.assertFalse('updatedtestvalue' in sm)
        self.assertTrue('oldvalue' in sm)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

import a10_neutron_lbaas.a10_exceptions as a10_ex

from a10_neutron_lbaas.tests.unit.v2 import fake_objs
from a10_neutron_lbaas.tests.unit.v2 import test_base


class TestLB(test_base.HandlerTestBase):

    def test_create(self):
        m = fake_objs.FakeLoadBalancer()
        self.a.lb.create(None, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.create' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('5.5.5.5' in s)
        self.assertTrue('UP' in s)

    def test_create_default_vrid_none_v21(self):
        self._test_create_default_vrid("2.1", None)

    def test_create_default_vrid_set_v21(self):
        self._test_create_default_vrid("2.1", 7)

    def test_create_default_vrid_none_v30(self):
        self._test_create_default_vrid("3.0", None)

    def test_create_default_vrid_set_v30(self):
        self._test_create_default_vrid("3.0", 7)

    def _test_create_default_vrid(self, api_ver=None, default_vrid=None):

        """
        Due to how the config is pulled in, we override the config
        for all of the devices.
        """

        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['default_virtual_server_vrid'] = default_vrid

        lb = fake_objs.FakeLoadBalancer()
        self.a.lb.create(None, lb)

        create = self.a.last_client.slb.virtual_server.create
        create.assert_has_calls([mock.ANY])
        calls = create.call_args_list

        if default_vrid is not None:
            self.assertIn('vrid=%s' % default_vrid, str(calls))
        if default_vrid is None:
            foundVrid = any(
                'vrid' in x.get('axapi_args', {}).get('virtual_server', {})
                for (_, x) in calls)
            self.assertFalse(
                foundVrid,
                'Expected to find no vrid in {0}'.format(str(calls)))

    def _test_create_template_virtual_server(self, api_ver=None, template_virtual_server=None):

        """
        Due to how the config is pulled in, we override the config
        for all of the devices.
        """

        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['template-virtual-server'] = template_virtual_server

        lb = fake_objs.FakeLoadBalancer()
        self.a.lb.create(None, lb)

        create = self.a.last_client.slb.virtual_server.create
        create.assert_has_calls([mock.ANY])
        calls = create.call_args_list

        if template_virtual_server is not None:
            self.assertIn("template_virtual_server='%s'" % template_virtual_server, str(calls))
        if template_virtual_server is None:
            foundVrid = any(
                'template_virtual_server' in x.get('axapi_args', {}).get('virtual_server', {})
                for (_, x) in calls)
            self.assertFalse(
                foundVrid,
                'Expected to find no template-virtual-server in {0}'.format(str(calls)))

    def test_create_template_virtual_server(self):
        self._test_create_template_virtual_server("2.1", None)

    def test_create_default_vrid_set_v21_with_template(self):
        self._test_create_template_virtual_server("2.1", "testTemplate")

    def test_create_default_vrid_none_v30_with_template(self):
        self._test_create_template_virtual_server("3.0", None)

    def test_create_default_vrid_set_v30_with_template(self):
        self._test_create_template_virtual_server("3.0", "testTemplate")

    def _test_create_virtual_server_templates(self, api_ver="3.0",
                                              virtual_server_templates=None, update=False):
        for k, v in self.a.config.get_devices().items():
            v['api_version'] = api_ver
            v['templates'] = virtual_server_templates

        lb = fake_objs.FakeLoadBalancer()
        if update:
            self.a.lb.update(None, lb, lb)
            create = self.a.last_client.slb.virtual_server.update
        else:
            self.a.lb.create(None, lb)
            create = self.a.last_client.slb.virtual_server.create
        create.assert_has_calls([mock.ANY])
        calls = create.call_args_list
        if virtual_server_templates is not None:
            self.assertIn('test-template-virtual-server', str(calls))
            self.assertIn('test-template-logging', str(calls))
            self.assertIn('test-policy', str(calls))
        if virtual_server_templates is None:
            foundVrid = any(
                'virtual_server_templates' in x.get('axapi_args', {}).get('virtual_server', {})
                for (_, x) in calls)
            self.assertFalse(
                foundVrid,
                'Expected to find no virtual_server_templates in {0}'.format(str(calls)))

    def test_create_with_template_virtual_server(self):
        template = {
            "virtual-server": {
                "template-virtual-server": "test-template-virtual-server",
                "template-logging": "test-template-logging",
                "template-policy": "test-policy",
                "template-scaleout": "test-scaleout",
            }
        }
        self._test_create_virtual_server_templates("3.0", virtual_server_templates=template)

    def test_update_with_template_virtual_server(self):
        template = {
            "virtual-server": {
                "template-virtual-server": "test-template-virtual-server",
                "template-logging": "test-template-logging",
                "template-policy": "test-policy",
                "template-scaleout": "test-scaleout",
            }
        }
        self._test_create_virtual_server_templates("3.0",
                                                   virtual_server_templates=template,
                                                   update=True)

    # There's no code that causes listeners to be added
    # if they are present when the pool is created.
    # We'd use unittest.skip if it worked with cursed 2.6

    # def test_create_with_listeners(self):
    #     pool = test_base.FakePool('HTTP', 'ROUND_ROBIN', None)
    #     m = test_base.FakeLoadBalancer()
    #     for x in [1, 2, 3]:
    #         z = test_base.FakeListener('TCP', 2222+x, pool=pool,
    #                                    loadbalancer=m)
    #         m.listeners.append(z)
    #     self.a.lb.create(None, m)
    #     s = str(self.a.last_client.mock_calls)
    #     print ("LAST CALLS {0}".format(s))
    #     self.assertTrue('call.slb.virtual_server.create' in s)
    #     self.assertTrue('fake-lb-id-001' in s)
    #     self.assertTrue('5.5.5.5' in s)
    #     self.assertTrue('UP' in s)
    #     self.assertTrue('vport.create' in s)
    #     for x in [1, 2, 3]:
    #         self.assertTrue(str(2222+x) in s)

    def test_update_down(self):
        m = fake_objs.FakeLoadBalancer()
        m.admin_state_up = False
        self.a.lb.update(None, m, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.update' in s)
        self.assertTrue('fake-lb-id-001' in s)
        self.assertTrue('5.5.5.5' in s)
        self.assertTrue('DOWN' in s)

    def test_delete(self):
        m = fake_objs.FakeLoadBalancer()
        self.a.lb.delete(None, m)
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.delete' in s)
        self.assertTrue(m.id in s)

    def test_delete_removes_slb(self):
        m = fake_objs.FakeLoadBalancer()
        self.a.lb.delete(None, m)

    def test_refresh(self):
        try:
            self.a.lb.refresh(None, fake_objs.FakeLoadBalancer())
        except a10_ex.UnsupportedFeature:
            pass

    def test_stats_v30(self):
        test_lb = fake_objs.FakeLoadBalancer()
        test_lb.stats_v30()
        c = mock.MagicMock()
        c.client.slb.virtual_server.get = mock.Mock(return_value=test_lb.virt_server)
        c.client.slb.service_group.stats = mock.Mock(return_value=test_lb.service_group)
        c.client.slb.service_group.get = mock.Mock(return_value=test_lb.members)
        ret_val = self.a.lb._stats_v30(c, test_lb.port_list, None)

        self.print_mocks()
        self.assertEqual(ret_val, test_lb.ret_stats_v30)

    def test_stats_v21(self):
        test_lb = fake_objs.FakeLoadBalancer()
        test_lb.stats_v21()
        c = mock.MagicMock()
        c.client.slb.virtual_service.get = mock.Mock(return_value=test_lb.virt_service)
        c.client.slb.service_group.stats = mock.Mock(return_value=test_lb.serv_group)
        ret_val = self.a.lb._stats_v21(c, test_lb.virt_server)

        self.print_mocks()
        self.assertEqual(ret_val, test_lb.ret_stats)

    def test_stats(self):
        test_lb = fake_objs.FakeLoadBalancer()
        self.a.lb.stats(None, test_lb)

        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('call.slb.virtual_server.stats' in s)

    def do_raise_exception(self, e, msg="mock raised exception"):
        def raise_exception(e, msg="acos broke!"):
            raise e(msg)

        return lambda *args, **kwargs: raise_exception(e, msg)

    def _test_create_expressions(self, os_name, pattern, expressions=None):
        self.a.config.get_virtual_server_expressions = self._get_expressions_mock
        expressions = expressions or self.a.config.get_virtual_server_expressions()
        expected = expressions.get(pattern, {}).get("json", None) or ""

        m = fake_objs.FakeLoadBalancer()
        m.name = os_name
        handler = self.a.lb
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("virtual_server.create", s)
        self.assertIn(str(expected), s)

    def test_create_expressions_none(self):
        self._test_create_expressions("mylb", None, {})

    def test_create_expressions_match_beginning(self):
        self._test_create_expressions("securelb", self.EXPR_BEGIN)

    def test_create_expressions_match_end(self):
        self._test_create_expressions("lbweb", self.EXPR_END)

    def test_create_expressions_match_charclass(self):
        self._test_create_expressions("lbwwlb", self.EXPR_CLASS)

    def test_create_expressions_nomatch(self):
        self.a.config.get_virtual_server_expressions = self._get_expressions_mock
        expressions = self.a.config.get_virtual_server_expressions()
        expected = expressions.get(self.EXPR_BEGIN, {}).get("json", None) or ""

        m = fake_objs.FakeLoadBalancer()
        m.name = "mylb"
        handler = self.a.lb
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("virtual_server.create", s)
        self.assertNotIn(str(expected), s)

    def test_create_noname_noexception(self):
        self.a.config.get_virtual_server_expressions = self._get_expressions_mock
        expressions = self.a.config.get_virtual_server_expressions()

        expected = expressions.get(self.EXPR_BEGIN, {}).get("json", None) or ""

        m = fake_objs.FakeLoadBalancer()
        m.name = None
        handler = self.a.lb
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("virtual_server.create", s)
        self.assertNotIn(str(expected), s)



# Copyright 2014-2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

vport_expressions = {
    "secure scaleout": {
        "regex": "^secure",
        "json": {
            "scaleout-bucket-count": 128
        }
    }
}

virtual_server_expressions = {
    "described": {
        "regex": "^desc",
        "json": {
            "description": "Described from the name"
        }
    }
}

service_group_expressions = {
    "mon1": {
        "regex": "^mon",
        "json": {
            "health-check": "mon1"
        }
    }
}

member_expressions = {
    "connections42": {
        "regex": "^connlimit",
        "json": {
            "conn-limit": 4200
        }
    }
}

monitor_expressions = {
    "monitor": {
        "regex": "^tagged",
        "json": {
            "user-tag": 42
        }
    }
}



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from a10_neutron_lbaas.tests.unit.v2 import fake_objs
from a10_neutron_lbaas.tests.unit.v2 import test_base


class TestHM(test_base.HandlerTestBase):

    def assert_hm(self, model, mon_type, method, url, expect_code, port):
        self.a.openstack_driver.health_monitor.successful_completion.assert_called_with(
            None, model)
        self.a.last_client.slb.hm.create.assert_called_with(
            'fake-hm-id-001', mon_type, 7, 7, 8,
            method=method, url=url, expect_code=expect_code,
            config_defaults=mock.ANY, port=port, axapi_args={},
        )

    def assert_create_sets_delay_timeout(self, model, mon_type, method, url, expect_code):
        model.timeout = 10
        model.delay = 6
        self.a.openstack_driver.health_monitor.successful_completion.assert_called_with(
            None, model)
        self.a.last_client.slb.hm.create.assert_called_with(
            'fake-hm-id-001', mon_type, model.delay, model.timeout, 8,
            method=method, url=url, expect_code=expect_code, axapi_args={})

    def test_create_ping(self):
        m = fake_objs.FakeHM('PING')
        self.a.hm.create(None, m)
        self.assert_hm(m, self.a.last_client.slb.hm.ICMP, None, None, None, mock.ANY)

    def test_create_tcp(self):
        m = fake_objs.FakeHM('TCP')
        self.a.hm.create(None, m)
        self.print_mocks()
        self.assert_hm(m, self.a.last_client.slb.hm.TCP, None, None, None, mock.ANY)

    def test_create_http(self):
        m = fake_objs.FakeHM('HTTP')
        self.a.hm.create(None, m)
        self.assert_hm(m, self.a.last_client.slb.hm.HTTP, 'GET', '/', '200', mock.ANY)

    def test_create_https(self):
        m = fake_objs.FakeHM('HTTPS')
        self.a.hm.create(None, m)
        self.assert_hm(m, self.a.last_client.slb.hm.HTTPS, 'GET', '/', '200', mock.ANY)

    def test_create_http_with_pool(self):
        m = fake_objs.FakeHM('HTTP', pool=mock.MagicMock())
        self.a.hm.create(None, m)
        self.assert_hm(m, self.a.last_client.slb.hm.HTTP, 'GET', '/', '200', mock.ANY)
        self.a.last_client.slb.service_group.update.assert_called_with(
            m.pool.id, health_monitor='fake-hm-id-001', health_check_disable=False)

    def test_update_tcp(self, m_old=None, m=None):
        if m_old is None:
            m_old = fake_objs.FakeHM('TCP')
        if m is None:
            m = fake_objs.FakeHM('TCP')
        m.delay = 20
        self.a.hm.update(None, m_old, m)
        self.a.openstack_driver.health_monitor.successful_completion.assert_called_with(
            None, m)
        self.a.last_client.slb.hm.update.assert_called_with(
            'fake-hm-id-001', self.a.last_client.slb.hm.TCP, 20, 7, 8,
            method=None, url=None, expect_code=None, port=None,
            config_defaults=mock.ANY, axapi_args={})

    def test_update_tcp_add_pool(self):
        m = fake_objs.FakeHM('TCP', pool=mock.MagicMock())
        self.test_update_tcp(m=m)
        self.print_mocks()
        self.a.last_client.slb.service_group.update.assert_called_with(
            m.pool.id, health_monitor='fake-hm-id-001', health_check_disable=False)

    def test_update_tcp_delete_pool(self):
        m_old = fake_objs.FakeHM('TCP', pool=mock.MagicMock())
        self.test_update_tcp(m_old=m_old)
        self.print_mocks()
        self.a.last_client.slb.service_group.update.assert_called_with(
            m_old.pool.id, health_monitor='', health_check_disable=True)

    def test_delete(self):
        m = fake_objs.FakeHM('HTTP')
        self.a.hm.tenant_id = "tenant-id"
        self.a.hm.delete(None, m)
        self.a.openstack_driver.health_monitor.successful_completion.assert_called_with(
            None, m, delete=True)
        self.a.last_client.slb.hm.delete.assert_called_with('fake-hm-id-001')

    def test_delete_with_pool(self):
        m = fake_objs.FakeHM('TCP', pool=mock.MagicMock())
        self.a.hm.delete(None, m)
        self.a.openstack_driver.health_monitor.successful_completion.assert_called_with(
            None, m, delete=True)
        self.a.last_client.slb.service_group.update.assert_called_with(
            m.pool.id, health_monitor='', health_check_disable=True)
        self.a.last_client.slb.hm.delete.assert_called_with('fake-hm-id-001')

    def _test_create_expressions(self, os_name, pattern, expressions=None):
        self.a.config.get_monitor_expressions = self._get_expressions_mock
        expressions = expressions or self.a.config.get_monitor_expressions()
        expected = expressions.get(pattern, {}).get("json", None) or ""

        m = fake_objs.FakeHM('HTTPS')
        m.name = os_name

        handler = self.a.hm
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("hm.create", s)
        self.assertIn(str(expected), s)

    def test_create_expressions_none(self):
        self._test_create_expressions("myhm", None, {})

    def test_create_expressions_match_beginning(self):
        self._test_create_expressions("securehm", self.EXPR_BEGIN)

    def test_create_expressions_match_end(self):
        self._test_create_expressions("hmweb", self.EXPR_END)

    def test_create_expressions_match_charclass(self):
        self._test_create_expressions("hmwwhm", self.EXPR_CLASS)

    def test_create_expressions_nomatch(self):
        self.a.config.get_monitor_expressions = self._get_expressions_mock
        expressions = self.a.config.get_monitor_expressions()
        expected = expressions.get(self.EXPR_BEGIN, {}).get("json", None) or ""

        m = fake_objs.FakeHM('HTTPS')
        m.name = "myhm"
        handler = self.a.hm
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("hm.create", s)
        self.assertNotIn(str(expected), s)

    def test_create_noname_noexception(self):
        self.a.config.get_monitor_expressions = self._get_expressions_mock
        expressions = self.a.config.get_monitor_expressions()
        expected = expressions.get(self.EXPR_BEGIN, {}).get("json", None) or ""

        m = fake_objs.FakeHM('HTTPS')
        m.name = None
        handler = self.a.hm
        handler.create(None, m)

        s = str(self.a.last_client.mock_calls)
        self.assertIn("hm.create", s)
        self.assertNotIn(str(expected), s)



# Copyright (c) 2018 A10 Networks
#
# Heavily inspired by networking-arista
# Source at https://github.com/openstack/networking-arista
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_log import log
from oslo_utils import uuidutils

# post-Kilo import.
try:
    from neutron.db.models.segment import NetworkSegment
except ImportError:
    # this moved to neutron.db.models.segment in kilo
    from neutron.plugins.ml2.models import NetworkSegment


from neutron.db import models_v2 as nmodels
from neutron.plugins.ml2.models import PortBinding
from neutron.plugins.ml2.models import PortBindingLevel

from a10_neutron_lbaas.plumbing.utils import IPHelpers
from acos_client import errors as acos_exc

LOG = log.getLogger(__name__)
# THIS FLAG IS CRITICAL FOR SYSTEMS WITH NO HPB
# Instead of checking the hierarchical tables, we just grab the seg id from the LB net
_HPB_TEST = True

# THIS FLAG IS CRUCIAL FOR KILO COMPATIBILITY
# THIS USES tenant_id INSTEAD OF project_id
_IS_KILO = False


class AcosWrapper(object):
    def __init__(self, client, *args, **kwargs):
        self._client = client

    def create_nat_pool(self, name, start_addr, end_addr, cidr):
        try:
            return self._client.nat.pool.create(name, start_addr, end_addr, cidr)
        except Exception as ex:
            raise ex

    def create_vlan(self, vlan_id, interfaces={}):
        try:
            return self._client.vlan.create(vlan_id, veth=True, **interfaces)
        except Exception as ex:
            LOG.debug(ex)
            pass

    def create_ve(self, ve_dict):
        try:
            return self._client.interface.ve.create(**ve_dict)
        # TODO(mdurrant) Narrow exception handling.
        except acos_exc.DhcpAcquireFailed:
            LOG.error("UNABLE TO CREATE VE INTERFACE {0}".format(str(ve_dict)))
            return None
        except Exception as ex:
            raise ex

    def get_ve(self, ve_ifnum):
        rv = {}
        try:
            ve = self._client.interface.ve.get_oper(ve_ifnum)
            rv = ve
        # TODO(mdurrant) Narrow exception handling
        except acos_exc.NotFound:
            # Eat the notfound error, return empty dict.
            pass
        except Exception as ex:
            raise ex
        return rv

    def update_vip(self, vip_id, mac_address, vlan_id):
        vip = None
        try:
            vip = self._client.slb.virtual_server.update(vip_id, mac=1,
                                                         mac_address=mac_address,
                                                         vlan=vlan_id)
        except Exception as ex:
            raise ex
        return vip


class NeutronDbWrapper(object):
    VLAN_PORT_NAME_FORMAT = "A10_VLANHBPHOOK_PROJECT_{project_id}_NET_{network_id}"
    VLAN_PORT_OWNER = "network:a10networks"
    BINDING_VNIC_TYPE = "normal"
    BINDING_PROFILE = ""
    BINDING_VIF_TYPE = "unbound"
    BINDING_VIF_DETAILS = ""

    """Wraps neutron DB ops for easy testing"""
    def __init__(self, session, *args, **kwargs):
        self._session = session

    def get_port(self, port_id):
        return self._session.query(nmodels.Port).filter(nmodels.Port.id == port_id).first()

    def get_segment(self, port_id, level):
        if _HPB_TEST:
            port = self._session.query(nmodels.Port).filter_by(id=port_id).first()
            segment = self._session.query(NetworkSegment).filter_by(
                network_id=port.network_id).first()
            return segment

        binding_level = self._session.query(PortBindingLevel).filter_by(
            port_id=port_id, level=level).first()
        if binding_level:
            segment = self._session.query(NetworkSegment).filter_by(
                id=binding_level.segment_id).first()
            return segment
        # No binding leve
        LOG.error("Could not find binding level for port:{0} level:{1}".format(port_id, level))

    def get_subnet(self, id):
        subnet = self._session.query(nmodels.Subnet).filter_by(id=id).first()
        return subnet

    def allocate_ip_for_subnet(self, subnet_id, mac, port_id):
        """Allocates an IP from the specified subnet and creates a port"""
        # Get an available IP and mark it as used before someone else does
        # If there's no IP, , log it and return an error
        # If we successfully get an IP, create a port with the specified MAC and device data
        # If port creation fails, deallocate the IP
        subnet = self.get_subnet(subnet_id)
        ip, mask, port_id = self.a10_allocate_ip_from_dhcp_range(subnet, "vlan", mac, port_id)
        return ip, mask, port_id

    def get_ipallocationpool_by_subnet_id(self, subnet_id):
        return self._session.query(nmodels.IPAllocationPool).filter(
            nmodels.IPAllocationPool.subnet_id == subnet_id).first()

    def get_ipallocations_by_subnet_id(self, subnet_id):
        return self._session.query(nmodels.IPAllocation).filter_by(subnet_id=subnet_id).all()

    def create_port(self, network_id, project_id, mac_address, device_id):
        device_owner = self.VLAN_PORT_OWNER
        device_id = network_id
        name = self.VLAN_PORT_NAME_FORMAT.format(project_id=project_id, network_id=network_id)
        port_dict = self._build_port_dict(project_id, name, network_id, mac_address, device_id,
                                          device_owner)
        return self.create_port_from_dict(port_dict)

    def create_port_from_dict(self, record):
        with self._session.begin(subtransactions=True):
            port = nmodels.Port(
                id=uuidutils.generate_uuid(),
                tenant_id=record['tenant_id'],
                name=record["name"],
                network_id=record["network_id"],
                mac_address=record["mac_address"],
                admin_state_up=record["admin_state_up"],
                # ACTIVE by default
                status=record["status"],
                device_id=record["device_id"],
                device_owner=record["device_owner"]
            )
            self._session.add(port)

            return port

    def get_ipallocation_with_port_by_subnet_id(self, subnet_id, ip_address, port_name):
        result = self._session.query(nmodels.IPAllocation, nmodels.Port)\
            .join(nmodels.Port, nmodels.Port.name == port_name)\
            .filter_by(nmodels.IPAllocation.ip_address == ip_address and
                       nmodels.IPAllocation.port_id == port_name and
                       subnet_id == nmodels.IPAllocation.subnet_id).first()
        return result

    def create_ipallocation(self, ipa):
        with self._session.begin(subtransactions=True):
            ipallocation = nmodels.IPAllocation(
                ip_address=ipa["ip_address"],
                network_id=ipa["network_id"],
                port_id=ipa["port_id"],
                subnet_id=ipa["subnet_id"]
            )
            self._session.add(ipallocation)

        return ipallocation

    def a10_allocate_ip_from_dhcp_range(self, subnet, interface_id, mac, port_id):
        """Search for an available IP.addr from unallocated nmodels.IPAllocationPool range.
        If no addresses are available then an error is raised. Returns the address as a string.
        This search is conducted by a difference of the nmodels.IPAllocationPool set_a
        and the current IP allocations.
        """
        subnet_id = subnet["id"]
        network_id = subnet["network_id"]

        iprange_result = self.get_ipallocationpool_by_subnet_id(subnet_id)
        ip_in_use_list = [x.ip_address for x in self.get_ipallocations_by_subnet_id(subnet_id)]

        range_begin, range_end = iprange_result.first_ip, iprange_result.last_ip
        ip_address = IPHelpers.find_unused_ip(range_begin, range_end, ip_in_use_list)

        if not ip_address:
            msg = "Cannot allocate from subnet {0}".format(subnet)
            LOG.error(msg)
            # TODO(mdurrant) - Raise neutron exception
            raise Exception

        mark_in_use = {
            "ip_address": ip_address,
            "network_id": network_id,
            "port_id": port_id,
            "subnet_id": subnet["id"]
        }

        self.create_ipallocation(mark_in_use)

        return ip_address, subnet["cidr"], mark_in_use["port_id"]

    def _build_port_dict(self, tenant_id, name, network_id, mac_address, device_id, device_owner):
        return {
            "id": uuidutils.generate_uuid(),
            "tenant_id": tenant_id,
            "name": name,
            "network_id": network_id,
            "mac_address": mac_address,
            "admin_state_up": 1,
            "status": "ACTIVE",
            "device_id": device_id,
            "device_owner": device_owner
        }

    def create_port_binding(self, port_id, host):
        return self._create_port_binding(port_id, host, self.BINDING_VNIC_TYPE,
                                         self.BINDING_PROFILE,
                                         self.BINDING_VIF_TYPE,
                                         self.BINDING_VIF_DETAILS)

    def _create_port_binding(self, port_id, host, vnic, profile, vif, vif_details, status="ACTIVE"):
        with self._session.begin(subtransactions=True):
            binding = PortBinding(
                port_id=port_id,
                host=host,
                vnic_type=vnic,
                vif_type=vif,
                profile=profile,
                vif_details=vif_details,
                # status=status,
            )
            self._session.add(binding)

    def update_port(self, port_id, mac):
        with self._session.begin(subtransactions=True):
            port = self._session.query(nmodels.Port).filter_by(id=port_id).first()
            port.mac_address = mac
            self._session.add(port)

    def cleanup_vlan_ports(self, last_lb):
        if hasattr(last_lb, "vip_port"):
            vip_port = last_lb.vip_port
            network_id = vip_port.network_id
            tenant_id = last_lb.tenant_id
        # v1 obj model
        else:
            # It's a pool, not a vip, we can work with that
            subnet_id = last_lb["subnet_id"]
            subnet = self.get_subnet(subnet_id)
            network_id = subnet["network_id"]
            tenant_id = last_lb["tenant_id"]

        # port_name = self.VLAN_PORT_NAME_FORMAT.format(project_id=tenant_id,network_id=network_id)
        project_id_fieldname = "project_id" if not _IS_KILO else "tenant_id"

        query_args = {
            "network_id": network_id,
            project_id_fieldname: tenant_id,
            # "name": port_name,
            "device_owner": self.VLAN_PORT_OWNER
        }

        with self._session.begin(subtransactions=True):
            port = self._session.query(nmodels.Port).filter_by(**query_args).first()
            if port:
                self._session.delete(port)



# Copyright 2019, Omkar Telee (omkartelee01), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

from a10_neutron_lbaas.v2 import handler_base_v2
from a10_neutron_lbaas.v2 import v2_context as a10

from a10_neutron_lbaas.v2.policy import PolicyUtil

LOG = logging.getLogger(__name__)


class L7PolicyHandler(handler_base_v2.HandlerBaseV2):
    def _set(self, set_method, c, context, file="", script="", size=0, action="", **kwargs):
        set_method(file=file, script=script, size=size, action=action)

    def create(self, context, l7policy, **kwargs):
        with a10.A10WriteStatusContext(self, context, l7policy) as c:
            try:
                filename = l7policy.id
                action = "import"
                p = PolicyUtil()
                script = p.createPolicy(l7policy)
                size = len(script.encode('utf-8'))
                self._set(c.client.slb.aflex_policy.create,
                          c, context, filename, script, size, action)
                old_listener = l7policy.listener
                new_listener = l7policy.listener
                get_listener = c.client.slb.virtual_server.vport.get(old_listener.loadbalancer_id,
                                                                     old_listener.name,
                                                                     old_listener.protocol,
                                                                     old_listener.protocol_port)

                if 'aflex-scripts' in get_listener['port']:
                    aflex_scripts = get_listener['port']['aflex-scripts']
                    aflex_scripts.append({"aflex": filename})
                    new_listener.aflex = aflex_scripts
                else:
                    new_listener.aflex = [{"aflex": filename}]
                self.a10_driver.listener.update(context, old_listener, new_listener)
            except acos_errors.Exists:
                pass

    def update(self, context, old_l7policy, l7policy, **kwargs):
        self.create(context, l7policy, **kwargs)

    def _delete(self, c, context, l7policy):
        LOG.debug("L7PolicyHandler.delete(): l7policy=%s, context=%s" % (l7policy.id, context))
        try:
            old_listener = l7policy.listener
            new_listener = l7policy.listener
            get_listener = c.client.slb.virtual_server.vport.get(old_listener.loadbalancer_id,
                                                                 old_listener.name,
                                                                 old_listener.protocol,
                                                                 old_listener.protocol_port)
            # removing listener attachment
            if 'aflex-scripts' in get_listener['port']:
                aflex_scripts = get_listener['port']['aflex-scripts']
                new_aflex_scripts = []
                for aflex in aflex_scripts:
                    if aflex['aflex'] != l7policy.id:
                        new_aflex_scripts.append(aflex)
                new_listener.aflex = new_aflex_scripts
            self.a10_driver.listener.update(context, old_listener, new_listener)
            c.client.slb.aflex_policy.delete(l7policy.id)
        except acos_errors.Exists:
                pass

    def delete(self, context, l7policy):
        with a10.A10DeleteContext(self, context, l7policy) as c:
            self._delete(c, context, l7policy)



# Copyright 2014-2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from netaddr import IPAddress as ipaddr
from netaddr import IPRange as iprange
from netaddr import IPSet as ipset

from oslo_log import log

import random

LOG = log.getLogger.__name__


class IPHelpers(object):
    @staticmethod
    def generate_random_mac(prefix):
        mac_address = [
            random.randint(0x00, 0xff),
            random.randint(0x00, 0xff),
            random.randint(0x00, 0xff)]
        mac_last = ':'.join(map(lambda x: "%02x" % x, mac_address))
        mac_address = "prefix:{0}".format(mac_last)
        return mac_address

    @staticmethod
    def find_unused_ip(ip_range_begin, ip_range_end, ips_in_use):
        candidate = None

        in_range = ipset(iprange(ip_range_begin, ip_range_end))
        in_use = ipset(ips_in_use)

        try:
            candidate = str(ipaddr((in_range - in_use).pop()))
        except Exception:
            LOG.error("Could not allocate IP address for range:{0}-{1}".format(
                ip_range_begin, ip_range_end))
        finally:
            return candidate



# Copyright 2019, Omkar Telee (omkartelee01), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from a10_neutron_lbaas.tests.unit.v2 import fake_objs
from a10_neutron_lbaas.tests.unit.v2 import test_base


class TestL7Policy(test_base.HandlerTestBase):
    def test_create_l7policy(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        m = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                   "http//:google.com", 23)
        self.a.l7policy.create(None, m)
        sm = str(self.a.last_client.mock_calls)
        self.assertTrue('fake-l7policy-id-001' in sm)

    def test_update_l7policy(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        m = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                   "http//:google.com", 23)
        m_new = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                       "http://youtube.com", 23)
        self.a.l7policy.update(None, m, m_new)
        sm = str(self.a.last_client.mock_calls)
        self.assertTrue('fake-l7policy-id-001' in sm)

    def test_delete_l7policy(self):
        lb = fake_objs.FakeLoadBalancer()
        flist = fake_objs.FakeListener('HTTP', 80, pool=None,
                                       loadbalancer=lb)
        flist.loadbalancer_id = "fake-lb-id-001"
        m = fake_objs.FakeL7Policy(flist, "REDIRECT_TO_URL", None,
                                   "http//:google.com", 23)
        self.a.l7policy.delete(None, m)
        sm = str(self.a.last_client.mock_calls)
        self.assertTrue('fake-listen-id-001' in sm)



# Copyright 2019, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from a10_neutron_lbaas.v2.policy import PolicyUtil

from a10_neutron_lbaas.tests.unit.v2 import fake_objs
from a10_neutron_lbaas.tests.unit.v2 import test_base


class TestPolicy(test_base.HandlerTestBase):
    def test_create_with_no_rules_action_REDIRECT_TO_URL(self):
        m = fake_objs.FakeL7Policy(None, "REDIRECT_TO_URL", None,
                                   "http://a10networks.com", 12)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """ when HTTP_REQUEST {
        if { ( true ) } {
        HTTP::redirect http://a10networks.com
        }
        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_no_rules_action_REDIRECT_TO_POOL(self):
        pool = fake_objs.FakePool('HTTP', 'ROUND_ROBIN', 'SOURCE_IP')
        m = fake_objs.FakeL7Policy(None, "REDIRECT_TO_POOL", pool,
                                   None, 12)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """  when HTTP_REQUEST {

        if { ( true ) } {

        pool fake-pool-id-001

        }

        } """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_no_rules_action_REJECT(self):
        m = fake_objs.FakeL7Policy(None, "REJECT", None,
                                   None, 12)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """ when HTTP_REQUEST {

        if { ( true ) } {

        HTTP::close

        }

        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_with_rule_action_REDIRECT_TO_URL(self):
        m = fake_objs.FakeL7Policy(None, "REDIRECT_TO_URL", None,
                                   "http://a10networks.com", 12)
        r = fake_objs.FakeL7Rule(m.id, 'HOST_NAME',
                                 'STARTS_WITH', None, 'testvalue')
        m.rules = []
        m.rules.append(r)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """ when HTTP_REQUEST {
        if { ([HTTP::host] starts_with "testvalue") } {
        HTTP::redirect http://a10networks.com
        }
        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_with_rule_action_REDIRECT_TO_POOL(self):
        pool = fake_objs.FakePool('HTTP', 'ROUND_ROBIN', 'SOURCE_IP')
        m = fake_objs.FakeL7Policy(None, "REDIRECT_TO_POOL", pool,
                                   None, 12)
        r = fake_objs.FakeL7Rule(m.id, 'FILE_TYPE',
                                 'STARTS_WITH', 'testkey', 'testvalue')
        m.rules = []
        m.rules.append(r)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """    when HTTP_REQUEST {
        if { ([HTTP::uri] ends_with "testvalue") } {
        pool fake-pool-id-001
        }
        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_with_rule_action_REJECT(self):
        m = fake_objs.FakeL7Policy(None, "REJECT", None,
                                   None, 12)
        r = fake_objs.FakeL7Rule(m.id, 'COOKIE',
                                 'CONTAINS', 'testkey', 'testvalue')
        m.rules = []
        m.rules.append(r)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """when HTTP_REQUEST {
        if { ([HTTP::cookie testkey] contains "testvalue") } {
        HTTP::close
        }
        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_with_multi_rule_action_REDIRECT_TO_URL(self):
        m = fake_objs.FakeL7Policy(None, "REDIRECT_TO_URL", None,
                                   "http://a10networks.com", 12)
        r1 = fake_objs.FakeL7Rule(m.id, 'HOST_NAME',
                                  'STARTS_WITH', None, 'testvalue')
        r2 = fake_objs.FakeL7Rule(m.id, 'FILE_TYPE',
                                  'STARTS_WITH', 'testkey', 'testvalue')
        r3 = fake_objs.FakeL7Rule(m.id, 'COOKIE',
                                  'CONTAINS', 'testkey', 'testvalue')
        m.rules = []
        m.rules.append(r1)
        m.rules.append(r2)
        m.rules.append(r3)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """  when HTTP_REQUEST {
        if { ([HTTP::host] starts_with "testvalue") and
        ([HTTP::uri] ends_with "testvalue") and
        ([HTTP::cookie testkey] contains "testvalue") } {
        HTTP::redirect http://a10networks.com
        }
        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)

    def test_create_with_with_rule_action_REDIRECT_TO_POOL_invert(self):
        pool = fake_objs.FakePool('HTTP', 'ROUND_ROBIN', 'SOURCE_IP')
        m = fake_objs.FakeL7Policy(None, "REDIRECT_TO_POOL", pool,
                                   None, 12)
        r = fake_objs.FakeL7Rule(m.id, 'FILE_TYPE',
                                 'STARTS_WITH', 'testkey', 'testvalue')
        r.invert = True
        m.rules = []
        m.rules.append(r)
        p = PolicyUtil()
        resp = p.createPolicy(m)
        expect_resp = """    when HTTP_REQUEST {
        if { not([HTTP::uri] ends_with "testvalue") } {
        pool fake-pool-id-001
        }
        }
        """
        resp = resp.replace('\n', '').replace(' ', '')
        expect_resp = expect_resp.replace('\n', '').replace(' ', '')
        self.assertEqual(resp, expect_resp)



# Copyright 2019, Omkar Telee (omkartelee01), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

from a10_neutron_lbaas.v2 import handler_base_v2
from a10_neutron_lbaas.v2 import v2_context as a10

from a10_neutron_lbaas.v2.policy import PolicyUtil

LOG = logging.getLogger(__name__)


class L7RuleHandler(handler_base_v2.HandlerBaseV2):

    def _set(self, set_method, c, context, file="", script="", size=0, action="", **kwargs):
        set_method(file=file, script=script, size=size, action=action)

    def create(self, context, l7rule, **kwargs):
        l7policy = l7rule.policy
        with a10.A10WriteStatusContext(self, context, l7policy) as c:
            try:
                filename = l7policy.id
                action = "import"
                p = PolicyUtil()
                script = p.createPolicy(l7policy)
                size = len(script.encode('utf-8'))
                self._set(c.client.slb.aflex_policy.create,
                          c, context, filename, script, size, action)
            except acos_errors.Exists:
                pass

    def update(self, context, old_l7rule, l7rule, **kwargs):
        self.create(context, l7rule, **kwargs)

    def delete(self, context, l7rule):
        with a10.A10DeleteContext(self, context, l7rule):
            policy = l7rule.policy
            rules = l7rule.policy.rules
            for index, rule in enumerate(rules):
                if rule.id == l7rule.id:
                    del rules[index]
                    break
            policy.rules = rules
            l7rule.policy = policy
            self.create(context, l7rule)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import sys
import types

import a10_neutron_lbaas.v1.v1_context as a10

from a10_neutron_lbaas.tests.unit.v1 import test_base


class FakeException(Exception):
    pass


class TestA10Context(test_base.UnitTestBase):

    def _py2_copy(self, func):
        copy_func = types.FunctionType(func.func_code, func.func_globals,
                                       name=func.func_name,
                                       argdefs=func.func_defaults,
                                       closure=func.func_closure)
        return copy_func

    def _py3_copy(self, func):
        copy_func = types.FunctionType(func.__code__,
                                       func.__globals__,
                                       name=func.__name__,
                                       argdefs=func.__defaults__,
                                       closure=func.__closure__)
        return copy_func

    def _copy_func(self, func):
        if sys.version_info[0] == 2:
            return self._py2_copy(func)
        else:
            return self._py3_copy(func)

    def _clean_import(self):
        a10.a10_context.A10Context.get_partition_key = self.partition_key_save

    def setUp(self, **kwargs):
        super(TestA10Context, self).setUp(**kwargs)
        self.handler = self.a.pool
        self.ctx = self._build_openstack_context()
        self.m = {'id': 'fake-id-001', 'tenant_id': 'faketen1'}
        self.partition_key_save = self._copy_func(a10.a10_context.A10Context.get_partition_key)
        self.addCleanup(self._clean_import)
        a10.a10_context.A10Context.get_partition_key = mock.Mock()
        a10.a10_context.A10Context.partition_key = 'faketen1'

    def test_context(self):
        with a10.A10Context(self.handler, self.ctx, self.m) as c:
            self.empty_mocks()
            c
        self.empty_close_mocks()

    def test_context_e(self):
        try:
            with a10.A10Context(self.handler, self.ctx, self.m) as c:
                self.empty_mocks()
                c
                raise FakeException()
        except FakeException:
            self.empty_close_mocks()

    def _set_api_version(self, api_version="2.1"):
        for k, v in self.a.config.get("devices").items():
            v['api_version'] = api_version

    def test_write_v21(self):
        self._set_api_version()
        with a10.A10WriteContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.last_client.system.action.activate_and_write.assert_called(None, "shared")
        self.a.last_client.session.close.assert_called_with()

    def test_write_v30(self):
        self._set_api_version("3.0")
        with a10.A10WriteContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.last_client.system.action.write_memory.assert_called()
        self.a.last_client.session.close.assert_called_with()

    def test_write_e(self):
        try:
            with a10.A10WriteContext(self.handler, self.ctx, self.m) as c:
                c
                raise FakeException()
        except FakeException:
            self.empty_close_mocks()
            pass

    def test_write_status(self):
        with a10.A10WriteStatusContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.openstack_driver._active.assert_called_with(
            self.ctx, 'pool', 'fake-id-001')

    def test_write_status_e(self):
        try:
            with a10.A10WriteStatusContext(self.handler, self.ctx,
                                           self.m) as c:
                c
                raise FakeException()
        except FakeException:
            self.a.openstack_driver._failed.assert_called_with(
                self.ctx, 'pool', 'fake-id-001')
            pass

    def test_delete(self):
        with a10.A10DeleteContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.openstack_driver._db_delete.assert_called_with(
            self.ctx, 'pool', 'fake-id-001')

    def test_delete_e(self):
        try:
            with a10.A10DeleteContext(self.handler, self.ctx, self.m) as c:
                c
                raise FakeException()
        except FakeException:
            self.empty_close_mocks()
            pass

    def test_partition_name(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            self.assertEqual(c.partition_name, 'shared')

    def test_partition_name_withalt(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-alt') as c:
            self.assertEqual(c.partition_name, 'mypart')


# Re-run all the context manager tests with appliance partitioning.
class TestA10ContextADP(TestA10Context):

    def setUp(self):
        super(TestA10ContextADP, self).setUp()
        self.reset_v_method('adp')

    def tearDown(self):
        self.reset_v_method('lsi')

    def reset_v_method(self, val):
        for k, v in self.a.config.get_devices().items():
            v['v_method'] = val

    def empty_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(1, len(self.a.last_client.mock_calls))
        self.a.last_client.system.partition.active.assert_called_with(
            self.m['tenant_id'])

    def empty_close_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(2, len(self.a.last_client.mock_calls))
        self.a.last_client.session.close.assert_called_with()

    def test_write_v21(self):
        self._set_api_version("2.1")
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            c
        self.a.last_client.system.action.activate_and_write.assert_called_with(
            mock.ANY)
        self.a.last_client.session.close.assert_called_with()

    def test_write_v30(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            c
        self.a.last_client.system.action.activate_and_write.assert_called_with(mock.ANY)
        self.a.last_client.session.close.assert_called_with()

    def test_write_v21_deleted_partition(self):
        self._set_api_version("2.1")
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            c
            c.partition_name = None

        self.a.last_client.system.action.activate_and_write.assert_called_with(None)
        self.a.last_client.session.close.assert_called_with()

    def test_write_v21_partition(self):
        self._set_api_version("2.1")
        expected = "part1"
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            c
            c.partition_name = expected

        self.a.last_client.system.action.activate_and_write.assert_called_with(expected)
        self.a.last_client.session.close.assert_called_with()

    def test_partition_name(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            self.assertEqual(c.partition_name, self.m['tenant_id'][0:13])

    def test_partition_name_withalt(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-alt') as c:
            # shared_partition has no effect on an ADP configured device
            self.assertEqual(c.partition_name, self.m['tenant_id'][0:13])


class TestA10ContextHA(TestA10Context):

    def setUp(self):
        super(TestA10ContextHA, self).setUp()

    def test_ha(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='ax4') as c:
            c
        self.a.last_client.ha.sync.assert_called_with('1.1.1.1', 'admin', 'a10')



# Copyright 2019, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.v2.v2_context as a10
import mock

from a10_neutron_lbaas.tests.unit.v2 import fake_objs
from a10_neutron_lbaas.tests.unit.v2 import test_base


class TestA10PartitionKey(test_base.UnitTestBase):
    def _build_openstack_context(self):
        admin_context = {
            "tenant_id": "admin"
        }

        return mock.Mock(get_admin_context=mock.Mock(return_value=admin_context))

    def setUp(self, **kwargs):
        super(TestA10PartitionKey, self).setUp(**kwargs)
        self.handler = self.a.pool
        self.ctx = self._build_openstack_context()
        self.m = fake_objs.FakeLoadBalancer()

    def test_use_parent(self):
        fake_keystone = mock.MagicMock()
        fake_keystone.client.projects.get = mock.MagicMock(
            return_value=fake_objs.FakeKeystoneClient("brick"))
        a10.a10_context.keystone_helpers.KeystoneFromContext = mock.MagicMock(
            return_value=fake_keystone)

        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            self.assertEqual(c.partition_key, "brick")

    def test_use_parent_no_parent(self):
        fake_keystone = mock.MagicMock()
        fake_keystone.client.projects.get = mock.MagicMock(
            return_value=fake_objs.FakeKeystoneClient())
        a10.a10_context.keystone_helpers.KeystoneFromContext = mock.MagicMock(
            return_value=fake_keystone)

        with a10.A10WriteContext(self.handler, self.ctx, self.m, device_name='axadp-noalt') as c:
            self.assertEqual(c.partition_key, "get-off-my-lawn")
